#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/StripChart.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Command.h>
#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <stdlib.h>
#include "tcpphone.h"

int ReadLine(int fd,char *buf,int size)
/* read(2) until a newline char is read, then replaces it with a '\0' */
{
  int cnt=0,tmp;
  do {
    tmp=read(fd,buf+cnt,1);
    cnt+=tmp;
  } while(tmp==1&&buf[cnt-1]!='\n'&&cnt<size);
  buf[cnt-tmp]='\0';
  if(cnt<=0) {
    close(fd);
    perror("ReadLine");
  }
  return cnt-tmp+1;
}

#define MAXGRAPHS 10

Widget graph,grLabel[MAXGRAPHS],scrol;
Display *dis;
Window  win;
GC      gc;
int    numGraphs,currentIdx=0,hSize=1000;
time_t refTime;
short  *history;
/* Short saves memory. history[(hSize-time+currentIdx)%hSize][graph] is */
/* the value of a certain variable, time seconds ago. -1 indicates that */
/* no value is available. */
typedef struct {
  Pixel  color[MAXGRAPHS];
  Pixel  background,grid,m10,v10;
} resType;
resType res;
typedef resType *resPtr;
static XtResource resList[]={
{"c0","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[0]),XtRImmediate,0},
{"c1","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[1]),XtRImmediate,0},
{"c2","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[2]),XtRImmediate,0},
{"c3","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[3]),XtRImmediate,0},
{"c4","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[4]),XtRImmediate,0},
{"c5","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[5]),XtRImmediate,0},
{"c6","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[6]),XtRImmediate,0},
{"c7","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[7]),XtRImmediate,0},
{"c8","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[8]),XtRImmediate,0},
{"c9","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,color[9]),XtRImmediate,0},
 /* Get background color from Widget */
{"grid","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,grid),XtRImmediate,0},
{"min10","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,m10),XtRImmediate,0},
{"vert10","C",XtRPixel,sizeof(Pixel),XtOffset(resPtr,v10),XtRImmediate,0},
};

void Draw(void)
{
  int i,start,j;
  short v1,v2;
  Dimension wid,hei;
  float top;
  
  XtVaGetValues(graph,XtNwidth,&wid,XtNheight,&hei,NULL);
  XtVaGetValues(scrol,XtNtopOfThumb,&top,NULL);
  start=(int)(top*hSize+0.5);
  if(start>hSize-wid) {
    start=hSize-wid;
    top=start/(float)hSize;
    XtVaSetValues(scrol,XtNtopOfThumb,*(XtArgVal *)&top,NULL);
  }
  
  XClearWindow(dis,win);
  start+=currentIdx;
  XSetForeground(dis, gc, res.v10);
  for(i=hei-10;i>0;i-=10) {
    XDrawLine(dis,win,gc,0,i,wid,i);
  }
  XSetForeground(dis, gc, res.grid);
  for(i=hei-100;i>0;i-=100) {
    XDrawLine(dis,win,gc,0,i,wid,i);
  }
  for(i=60-(start+refTime)%60;i<wid;i+=60) {
    XDrawLine(dis,win,gc,i,0,i,hei);
  }
  XSetForeground(dis, gc, res.m10);
  for(i=600-(start+refTime)%600;i<wid;i+=600) {
    XDrawLine(dis,win,gc,i,0,i,hei);
  }
  for(i=0;i<numGraphs;i++) {
    XSetForeground(dis, gc, res.color[i]
    #ifdef XOR_DRAW
      ^res.background
    #endif
    );
    v2=history[start%hSize*numGraphs+i];
    for(j=1;j<wid;j++) {
      v1=v2;
      v2=history[(start+j)%hSize*numGraphs+i];
      if(v1>=0&&v2>=0) {
        XDrawLine(dis,win,gc,j
        #ifdef XOR_DRAW
          +1
        #endif
        ,hei-1-v1,j+1,hei-1-v2);
      }
    }
  }
}

void ReadWriter(XtPointer cliData,int *fid,XtInputId *id)
{
  int i;
  char buf[80],*ptr;
  
  if(ReadLine(*fid,buf,sizeof(buf))<=1) exit(1);
  /* If we only get a 0 terminator, then exit */
  ptr=strtok(buf,",\n");
  i=0;
  while(ptr!=NULL) {
    XtVaSetValues(grLabel[i],
      XtNlabel,ptr,NULL);
    if(sscanf(ptr,"%hd",&history[currentIdx*numGraphs+i])!=1) {
      history[currentIdx*numGraphs+i]=-1;
    }
    i++;
    ptr=strtok(NULL,",\n");
  }
  currentIdx++;
  if(currentIdx>=hSize) {
    currentIdx=0;
    refTime+=hSize;
  }
  
  Draw();
}

void Quit(Widget w, XtPointer cliData, XtPointer abc)
{
  exit(0);
}

int main(int argc,char *argv[])
{
  Widget top,box,quit,pane;
  XFontStruct font;
  XtAppContext ac;
  int works=1,server;
  float shown;
  Dimension tmp;
  XtArgVal *l_top;
  char *service="UPSstats",buf[80],*ptr;

  time(&refTime);
  refTime-=hSize;
  
  if(argc>=3) service=argv[2];
  top = XtVaAppInitialize(&ac, "Graph", NULL, 0, &argc, argv, NULL, NULL);
  XtGetApplicationResources(top,&res,resList,
    XtNumber (resList),NULL, 0);
  switch(argc) {
    case 4:works=works&&sscanf(argv[3],"%d",&hSize)==1;
    case 2:works=works&&isalnum(argv[1][0]);
    case 1:
      break;
    default:
      works=0;
      break;
  }
  if(!works) {
    fprintf(stderr,"Usage : %s [machinename [portNo [hSize]]]\n",
      argv[0]);
    return 1;
  }
  if((server=TCPPhone(argc>1?argv[1]:"localhost",service))<0) return 2;
  
  pane=XtVaCreateManagedWidget ("mainWindow",panedWidgetClass,top,NULL);
  
  graph = XtVaCreateManagedWidget("graph", simpleWidgetClass, pane, NULL);
  box=XtVaCreateManagedWidget ("box",boxWidgetClass,pane,NULL);

  XtVaGetValues(graph,XtNbackground,&res.background,NULL);  
  
  XtVaGetValues(graph,XtNwidth,&tmp,NULL);
  shown=tmp/(float)hSize;
  l_top=(XtArgVal*)&shown;
  scrol =XtVaCreateManagedWidget("scrol", scrollbarWidgetClass, box,
    XtNorientation,XtorientHorizontal,
    XtNwidth,tmp,
    XtNshown,*l_top,
    NULL);
  shown=1-shown;
  XtVaSetValues(scrol,
    XtNtopOfThumb,*(XtArgVal *)&shown,
    NULL);

  ReadLine(server,buf,sizeof(buf));
  ptr=strtok(buf,",\n");
  numGraphs=0;
  while(ptr!=NULL) {
    grLabel[numGraphs]=XtCreateManagedWidget(ptr,labelWidgetClass, box,
      NULL,0);
    XtVaSetValues(grLabel[numGraphs],
      XtNforeground,res.color[numGraphs],NULL);
    numGraphs++;
    ptr=strtok(NULL,",\n");
  }
  history=(short*)malloc(numGraphs*hSize*sizeof(*history));
  memset(history,-1,numGraphs*hSize*sizeof(*history));
  quit=XtCreateManagedWidget ("Quit",commandWidgetClass,box, NULL,0);
  XtAddCallback(quit,XtNcallback, Quit, NULL);
  
  XtAppAddInput(ac,server,(XtPointer)XtInputReadMask,ReadWriter,NULL);
  XtRealizeWidget(top);

  dis=XtDisplay(graph);
  win=XtWindow(graph);
  gc=XCreateGC(dis,win,0,NULL);
  #ifdef XOR_DRAW
  XSetFunction(dis,gc,GXxor);
  /* XOR makes it possible to see two graphs with the same value */
  #endif
  XtAppMainLoop(ac);
  return 0;
}

