/* (c) Nic Roets (geo@pta.lia.net), GPL2 applies. */
/* The "Safeware" software and the "Poweralert" software/Tripplite UPS is
described. After that the program that implements both is given. */

/* The "Safeware" software uses the following protocol to communicate with
the UPS (this is what I saw with a protocol analyzer and a digitizing
ossiliscope) :
Signal levels	: RS232
Mode		: Asynchrones, Halfduplex
Baudrate	: 1200
Bit sense	: Normal
Byte size	: 8
Parity		: None
Stop bits	: TWO ! (Weird but Important !)
RTS		: On for the first 250 seconds, then off.
DTR		: On
All messages begin with a 0x28 (open braket '(' in ASCII).
All messages end with a 0x29 (close braket ')' in ASCII).
The UPS only sends a message after receiving a request from the PC.
All bytes are seen as unsigned.

When the PC software starts up:
  PC takes RTS high for 250 ms, then wait for 210ms before sending :
  PC:0x28,0x4F,0x4B,0x29			(OK)
  UPS:0x28,0x54,0x4D,0x53,0x29			(TMS)

Every second the PC software is running:
  PC:0x28,0x44,0x10,0x29			(D\x10)
  UPS:0x28,mode,0x41,0x1F,inVolts,inFreq,outVolts,outFreq,batt,ld,0x29
  Mode is 0x84 for normal or 0x85 for power failure.0x8c for test.
  When used in 250V region, the input voltage was inVolts*2, and the output
    voltage was outVolts*2.
  The frequecies was inFreq and outFreq.
  The battery is about batt*100/255 % full.
  The load is about ld*100/80 %. Note that the reported load increases when
    the power fails, potentially causing overloads.
    
To set the wakeup time of the UPS:
  PC:0x28,0x57,0,MSB,LSB,0x29			(W\0  )
  UPS:???					(TMS)
  The shortest time that the UPS will disable the power is MSB*256+LSB
  minutes

Shutdown command:
  PC:0x28,0x53,sec,0x29				(S )
  UPS:0x28,0x54,0x4D,0x53,0x29			(TMS)
  The UPS will for sec seconds and then remove power, and then
  wait some more, and then wake up again. The version of the "Safeware"
  program I had, claimed a value of more than 255 is valid, but the analizer
  showed that "Safeware" DOES NOT SEND ANYTHING if more than 256 seconds is
  specified.
*/
/* This is the protocol for a "Smart 2200 rmint..." from Tripp Lite
(The power people) Made in USA, Win95, sorry I mean Chicago Il.
The format is an ASCII char (Non printables are replaced with a dot),
followed by the hex, followed by a space. When the data is ascii I
delete the hex codes. All messages are started with 1 oolon, and then a
capital which indicate the command.

Baudrate : 2400
Databits : 8
Parity   : None
Stop bits: 1

Computer : : W .0d
UPS      : : W .0d
Computer : : W .0d
UPS      : : W .0d .0a 838 D44 .0d .0a
Computer : : L .0d
UPS      : : L .0d .0a C 5 .0d .0a
Computer : : V .0d
UPS      : : V .0d .0a 030 838 .0d .0a
Computer : : D .0d
UPS      : : D .0d .0a 1 1 7 C 0 0 C 6 4 0 0 0 0 7 5 8 1 1 F 3 .0d .0a
Computer : : B .0d
UPS      : : B .0d .0a 8 A .0d .0a
Computer : : C .0d
UPS      : : C .0d .0a 0 0 .0d .0a

The pattern then repeats with commands D, B and C.

The temperature was 75.4 F (24.4 C)
With recepticles on :
UPS: : D . . 1 1 7 B 0 2 7 D 6 4 0 3 0 0 7 5 8 1 1 F 4 . .
     : B . . 8 A . .
     : C . . 0 0 . .

With rec on and Battery on :
UPS: : D . . 1 1 7 B 0 3 7 D 6 4 0 0 0 0 7 5 8 1 1 F 4 . .

When the temperature was 70.2 F (21.2 C)
UPS: : D . . 1 1 7 9 0 2 7 4 6 4 0 0 0 0 0 0 8 1 1 F 4 . .

Turn recepticles off:
Computer: : K 0 .
UPS echos
Later...
UPS: : D . . 1 1 7 B 0 0 7 D 6 4 0 0 0 0 7 5 8 1 1 F 3 . .

Turn recepticles on:
Computer: : K 1 .
UPS echos
Later...
UPS: : D . . 1 1 7 B 0 2 7 D 6 4 0 0 0 0 7 5 8 1 1 F 3 . .

Inverter self test:
Computer: : A .
UPS echos
Later...
UPS: : D . . 1 1 7 B 0 3 7 D 6 4 0 0 0 0 7 5 8 1 1 F 4 . .

Shut off inverter with 23 seconds delay:
Computer: : N 1 7 .
Computer: : G .
Computer: : K 0 .

UPS echos
Later...
UPS: : D . . 1 1 7 C 0 2 7 D 6 4 0 0 0 0 7 5 8 1 1 F 4 . .

To reboot after 63 seconds, send : H 0 0 0 0 3 F .
To enable remote reboot, send : Q 1 .
To enable remote power on, send : Q 0.

When the power fails :
UPS: : D . . 1 1 3 0 0 0 3 7 9 6 4 0 0 0 0 0 0 8 1 1 F 4
     : B 7 C

Shutdown OS after 60 seconds, and UPS after 120 seconds.
Therefore grace period is 60 seconds. :
Computer : : N 4 B .
Computer : : G .
Computer : : K 0 .
Note that 0x4B = 60 + 15, so the program adds a little bit, just for safety.
When the power fails :
UPS: : D . . 1 1 3 0 0 0 3 7 C 6 4 0 0 0 0 0 0 8 1 1 F 4
     : B 7 E

Shutdown OS after 50 seconds, and UPS after 70 seconds.
Therefore grace period is 20 seconds. :

Computer : : N 2 3 .
Computer : : G .
Computer : : K 0 .
If the last command (K 0) is not sent, then the UPS will automatically apply
power to the outputs when the power is returned.

When the power fails :
UPS: : D . . 1 1 3 0 0 0 3 7 C 6 4 0 0 0 0 0 0 8 1 1 F 4
     : B 7 E

If the grace period is more than 255 seconds, the computer sends N F F,
and the UPS will then only give the computer 4 and a quater minutes to
shutdown ! I analized 2 UPS programs and BOTH make the mistake of trying
to fit a word in a byte !!!

Without load : : D . . 1 1 7 A 0 2 7 D 6 4 0 0 0 0 0 0 8 1 1 F 4
With 2% load : : D . . 1 1 7 A 0 2 7 D 6 4 0 0 0 2 0 0 8 1 1 F 4

*/
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <stdio.h>
#include <string.h>
#include "powerdChunk.h"

extern void wall(); /* Belongs in a lib */

const char *writerServiceName="UPSstats";

static char message[81]="UPS not configured\n";
static char upsId[20]; /* Not used */

enum { invalid,unknown,unconfigured,configured,goingDown
} static state=invalid;

static int port=-1,syncSec,downMinutes,normalVolts,warnSec=300;
enum { safeWareProto,trippLiteProto } static proto;
static nextCommand='D'; /* Cycle throu trippLite commands. */
static unsigned long shutdownTime=0xFFFFFFFF,lastTime;
/* With this, I am obviously planning to cripple the worlds computers in */
/* the year 2106 */
static struct termios portSettings;
/* Too many static variables. */

char confLine[80];

int  PortRead(unsigned char *buf,int size)
{
  int cnt;
  struct timeval tv,tv2;
  struct timezone tz;
  
  gettimeofday(&tv,&tz);
  cnt=0;
  do {
    cnt+=read(port,buf+cnt,size-cnt);
    gettimeofday(&tv2,&tz);
  } while((tv2.tv_sec-tv.tv_sec)*1000000l+tv2.tv_usec-tv.tv_usec < 300000l&&
  /* Keep reading for 300ms */
  cnt<size);
  
  return cnt;
}

void TurnUpsOff(int signum) /* and tell init */
{ 
  char buf[10];
  int  cnt;

  powerfail(0);
  if (proto==safeWareProto) {
    do {
      buf[0]='(';
      buf[1]='S';
      buf[2]=syncSec;
      buf[3]=')';
      write(port,buf,4);
      cnt=PortRead(buf,sizeof(buf));
    } while(cnt < 2 || buf[0] != '(' || buf[cnt-1] !=')');
  }
  else {
    do {
      sprintf(buf,":N%02x\r",syncSec<256?syncSec:255);
      write(port,buf,5);
    } while(PortRead(buf,sizeof(buf)) < 5 && buf[0] != ':');
    do {
      write(port,":G\r",3);
    } while(PortRead(buf,sizeof(buf)) < 3 && buf[0] != ':');
    #ifdef NO_AUTO_RESTART
    do {
      write(port,":K0\r",4);
    } while(PortRead(buf,sizeof(buf)) < 4 && buf[0] != ':');      
    #endif
  }
  state=goingDown;
}

void Configure(void)
{
  FILE *confFile;
  char *ptr;
  static setSignal=0;
  
  if(!setSignal) {
    signal(SIGTERM,SIG_IGN); /* Incase shutdown process send us one */
    signal(SIGUSR1,TurnUpsOff);
    setSignal=1;
  }
  
  if(port>=0) close(port);
  port=-1;
  state=invalid;
  strcpy(message,
  "smartups: Please edit /etc/power, and \"kill -HUP\" me to reread it\n");
  if((confFile=fopen("/etc/power","r"))!=NULL) {
    fgets(confLine,sizeof(confLine),confFile);
    ptr=strtok(confLine," ,\n");
    if(ptr!=NULL) {
      if((port=open(ptr,O_RDWR|O_NOCTTY))==-1) {      
        sprintf(message,"Unable to open %s\n",ptr);
      }
      else {
        ptr=strtok(NULL," ,\n");
        if(ptr!=NULL) {
          syncSec=atoi(ptr);
          ptr=strtok(NULL," ,\n");
          if(1<=syncSec&&syncSec<=255&&ptr!=NULL) {
            downMinutes=atoi(ptr);
            ptr=strtok(NULL," ,\n");
            if(1<=downMinutes&&downMinutes<=255&&ptr!=NULL) {
              normalVolts=atoi(ptr); /* Do not care what it is */
              ptr=strtok(NULL," ,\n");
              if(ptr!=NULL && (warnSec=atoi(ptr)) >= 0) {
                tcgetattr(port,&portSettings);
                portSettings.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|
                  INLCR|IGNCR|ICRNL|IXON);
                portSettings.c_oflag &= ~OPOST;
                portSettings.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
                portSettings.c_cflag &= ~(CSIZE|PARENB);
                portSettings.c_cflag = CS8;
                portSettings.c_cc[VMIN]=0;
                portSettings.c_cc[VTIME]=3;
  
                state=unknown;
              }
            }
          }
        }
      }
    }
  }
}

void CheckShutdownState(int supplyOn)
{
/* Print the first 10 chars of message. */
  strcpy(message,supplyOn ? "Supply On " : "Supply Off");
  if(supplyOn) {
    if (shutdownTime<0xFFFFFFFF) {
      wall("Power is back\n",1);
      shutdownTime=0xFFFFFFFF;
    }
  }
  else {
    if(shutdownTime==0xFFFFFFFF) {
      time(&lastTime);
      shutdownTime=lastTime+warnSec;
    }
  }
}

static unsigned char buf[30];

int TalkWithTripp(const char *out,int expectLen)
/* The TrippLite sometimes gives a short answer, as if it is confused, so */
/* we allow two attempts with a delay inbetween. */
{
  int cnt,len;

  for(cnt=0;cnt<2;cnt++) {
    write(port,out,strlen(out));
    len = PortRead(buf,sizeof (buf));
    if(len>=expectLen && buf[0]==':' && buf[1]==out[1] && buf[2]==out[2]) {
      return len;
    }
    usleep(300000);
  }
  return 0;
}

const char *Monitor(void)
{
  int cnt,vin,hzin,vout,hzout,batt,load,tmp,i,degrees;
  char downMsg[]="(W\0  )",wallStr[160];
  long now;
  static oldBatt=100;
  
  #ifdef UPSLESS_TEST
  static cntDown=20;
  printf("tick %d\n",cntDown);
  if(--cntDown<=0) {
    CheckShutdownState(0);
  }
  #endif
  
  if(port>=0) {
    if(state==unknown) {
      portSettings.c_cflag |= CSTOPB;
      cfsetispeed(&portSettings,B1200);
      cfsetospeed(&portSettings,B1200);
      tcsetattr(port,TCSADRAIN,&portSettings);
  
      ioctl(port,TIOCMGET,&tmp);
      tmp|=TIOCM_RTS;
      ioctl(port,TIOCMSET,&tmp);
      usleep(250*1000);
      tmp&=~TIOCM_RTS;
      ioctl(port,TIOCMSET,&tmp);
      usleep(210*1000);
      write(port,"(OK)",4);
  
      cnt=PortRead(upsId,sizeof(upsId)-1);
      if(cnt > 4 && upsId[0] == '(' && upsId[cnt-1] == ')') {
      /* Previously the program allowed an id of any length to be accepted */
      /* but some devices just echo the (OK) we sent, so we have to check */
      /* for something extra like length. */
        upsId[cnt]='\0';
        state=unconfigured;
        proto=safeWareProto;
      } else {
        /* Now let's test for a TrippLite */
        portSettings.c_cflag &= ~CSTOPB;
        cfsetispeed(&portSettings,B2400);
        cfsetospeed(&portSettings,B2400);
        tcsetattr(port,TCSADRAIN,&portSettings);
        
        for(i=0;;i++) {
          if(i >= 2) { /* Is two retries always enough ? */
            Configure (); /* Closing and reopening the port is nessary for */
            /* some reason (when recovering when the talking to the tripp. */
            sprintf(message,"No UPS found on %s\n",confLine);
            break;
          }
          write(port,":W\r",3);
          cnt=PortRead(buf,sizeof(buf)-1);
          if(cnt>3&&memcmp(buf,":W\r\n",4)==0) {
            sprintf(upsId,"TrippLite %c",buf[4]);
            write(port,":L\r",3);
            cnt=PortRead(buf,sizeof(buf)-1);
            if(cnt>3&&memcmp(buf,":L\r\n",4)==0) {
              write(port,":V\r",3);
              cnt=PortRead(buf,sizeof(buf)-1);
              if(cnt>3&&memcmp(buf,":V\r\n",4)==0) {
                state=configured;
                /* No autorestart parameter to set ? Start to polling */
                proto=trippLiteProto;
                break;
              }
            }
          } /* If initial contact was made */
        } /* For each retry */
      }
      #ifdef DEBUG
      printf("%d %.*s\n",cnt,cnt,upsId);
      #endif
    }
    if(state==unconfigured) { /* Does not apply to trippLite */
      downMsg[3]=(char)(downMinutes>>8);
      downMsg[4]=(char)downMinutes;
      write(port,downMsg,6);
      cnt=PortRead(buf,sizeof(buf));
      #ifdef DEBUG
      printf("%d %.*s\n",cnt,cnt,buf);
      #endif
      if(cnt > 0 && buf[0]=='(' && buf[cnt-1]==')') {
        state=configured;
        #ifdef DEBUG
        printf("Configured\n");
        #endif
      }
      else state=unknown;
    }
    if(state==configured && proto==safeWareProto) {
      write(port,"(D\x10)",4);
      if((cnt=PortRead(buf,sizeof(buf)))>=11&&buf[0]=='('&&buf[cnt-1]==')') {
        vin=buf[4]*2;
        hzin=buf[5];
        vout=buf[6]*2;
        hzout=buf[7];
        batt=(buf[8]-175)*100/80;
        load=buf[9]*0.76865+10.78;
        if ((buf[1]|1) == 0x85) { /* 0x85 and 0x84 */
          CheckShutdownState(buf[1]==0x84);
        }
        else if (buf[1]==0x94) {
          strcpy(message, "Load & Sil");
        }
        else if (buf[1]==0x8C) {
          strcpy(message, "Test mode ");
        }
        else {
          sprintf (message,"Mode %3x ?",buf[1]);
        }
        sprintf (message+10,
        ",%3d Vin,%2d Hz in,%3d Vout,%2d Hz out,%3d%% batt,%3d%% load\n",
        vin,hzin,vout,hzout,batt,load);
      }
      else {
        sprintf(message,"No UPS on %s\n",confLine);
        state=unknown;
      }
      #ifdef DEBUG
      printf("%d %.*s\n",cnt,cnt,buf);
      #endif
    } /* if it is configured for safeWareProto */
    if (state == configured && proto== trippLiteProto) {
      if(nextCommand=='D') {
        if(TalkWithTripp(":D\r",26)) {
        /* Scan the string back to front, so that '\0' can be added as */
        /* required. */
          buf[18]='\0';
          sscanf(buf+16,"%x",&load);
          buf[12]='\0';
          sscanf(buf+10,"%x",&degrees);
          buf[8]='\0';
          sscanf(buf+6,"%x",&vin);
          vin=vin*2-10;
          /* Frequency ? PowerAlert lied to me and said I have 60 Hz mains */
          /* (South-Africa has 50 Hz), so I doubt wheather it can be */
          /* measured by the Smart UPS 2200 from TrippLite. */
          /* After I installed an update it was correct, but the */
          /* measurements only have a resolution of 1 Hz, which is not */
          /* small enough for me to see anything. */
          CheckShutdownState(vin > 80);
          /* (buf[6]-'0')&4 have some special meaning, but it is NOT */
          /* supply off. */
          sprintf(message+10,",%3d Vin,%3d%% batt,%3d%% load,%3.1lf %c\n",
          vin > 0 ? vin : 0, oldBatt,load,
          #ifndef CELCIUS
          degrees/1.65,'F');
          #else
          (degrees/1.65-32)/2.12,'C');
          #endif
        }
        else {
          sprintf(message,"No UPS on %s\n",confLine);
          state=unknown;
        }
        nextCommand='B';
      } else if(nextCommand=='B') {
        if(TalkWithTripp(":B\r",8)) {
          buf[6]='\0';
          sscanf(buf+4,"%x",&batt);
          oldBatt=batt<0x6E?10:(
                    batt<0x7A?(batt-0x6E)*5+20:(
                      batt<0x7D?80:(
                        batt<0x84?90:100
                      )
                    )
                  );
        }
        nextCommand='C';
        /* Same message */
      } else if(nextCommand=='C') {
        TalkWithTripp(":C\r",3);
        nextCommand='D';
      }
    }
    if (state == goingDown) {
      strcpy(message, "Ups WILL cut the power\n");
    }
  } /* if we have a port */
  if(shutdownTime<0xFFFFFFFF) {
    time(&now);
    sprintf(wallStr,"%sShutdown begins in %3ld seconds.",message,
      shutdownTime-now);
    if(  (shutdownTime-now<300 && shutdownTime-lastTime>=300) ||
         (shutdownTime-now<60  && shutdownTime-lastTime>=60)) {
      wall(wallStr,1);
    }
    if (shutdownTime<=now && shutdownTime > lastTime) TurnUpsOff(0);
    lastTime=now;
    if (shutdownTime-now<5) {
      state=goingDown;
    }
  }
  #ifdef DEBUG
  printf("%s",message);
  #endif
  return message;
}

