/* (c) Nic Roets (nic@geo.asd.co.za), GPL2 applies. */
/* To do : Find a solution for the "Address already in use" problem */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include "writerDaemon.h"

static configured;

void Unconfigure(int signum)
{
  configured=0;
}

void main(void)
{
  int lsock,len,startup=1,tmp;
  struct sockaddr_in upsAddr, peerAddr;
  struct servent *sp;
  typedef struct _client {
    struct _client *next;
    int fd;
  } client;
  client *clientList=NULL,*tmpClient,**clientRemover;
  fd_set set;
  struct timeval lastTime,tv;
  struct timezone tz;
  const char *ptr;

  signal(SIGPIPE,SIG_IGN);
  FD_ZERO(&set);
  
  if ((lsock=socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
    perror("socket");
    exit(1); /* This should never happen */
  }
  upsAddr.sin_family = AF_INET;
  upsAddr.sin_addr.s_addr=INADDR_ANY;
  
  if ((sp=getservbyname (writerServiceName, "tcp"))==NULL) {
    fprintf(stderr,"Please make an entry for %s in"
    " /etc/services\n", writerServiceName);
    exit(2);
  }
  upsAddr.sin_port = sp->s_port;
  
  if(bind(lsock,(struct sockaddr *)&upsAddr,sizeof(upsAddr))==-1) {
    perror("bind");
    exit(3);
  }

  if(listen (lsock, 1)==-1) {
    perror("listen");
    exit(4);
  }
                      
  gettimeofday(&lastTime,&tz);
  /* We will assume, for simplicity that lastTime.tv_usec is always 0 */
  while(1) {
    Configure();
    configured=1;
    if(signal(SIGHUP,Unconfigure)==SIG_ERR) {
      #if DEBUG
      perror("HUP");
      #endif
    }
    while(configured) {
      ptr=Monitor();
      if(startup) {
        printf("%s",ptr); /* Tell the user what is going on */
        #ifndef DEBUG
        switch(fork()) {
        case 0:
          setsid(); /* Become a child of init */
          break;
        case -1:
          perror("fork");
          exit(1);
        default:
          exit(0);
        } /* Un*x quiz : Why the fork ? So that the shell get a signal */
          /*             that it can continue when we exit. */
        #endif
        startup=0;
      }
      len=strlen(ptr);
      for(clientRemover=&clientList;*clientRemover!=NULL;) {
        tmpClient=*clientRemover;
        /*if(write(tmpClient->fd,ctime((time_t*)&lastTime.tv_sec),25)!=25) {*/
        if(write(tmpClient->fd,ptr,len)<=0 && errno!=EAGAIN) {
          #if DEBUG
          perror("Some client");
          #endif
          close(tmpClient->fd);
          *clientRemover=tmpClient->next;
          /* clientRemover stays the same ! */
          free(tmpClient);
        }
        else clientRemover=&tmpClient->next;
      }
      gettimeofday(&tv,&tz);
      lastTime.tv_sec++; /* Calculate the next time */
      if(tv.tv_sec<lastTime.tv_sec) tv.tv_usec=999999-tv.tv_usec;
      else tv.tv_usec=0; /* If it is already time, then hurry ! */
      tv.tv_sec=0; /* Always wait less than a second */
      
      /*FD_ZERO(&set);*/
      FD_SET (lsock, &set);
      if (select (lsock+1, &set, NULL, NULL, &tv)==-1) {
        #ifdef DEBUG
        perror("select");
        #endif
      }
      else if (FD_ISSET (lsock, &set)) {
        len=sizeof (peerAddr);
        tmpClient=(client *)malloc(sizeof(*tmpClient));
        if((tmpClient->fd=accept(lsock, (struct sockaddr *)&peerAddr, &len))==-1) {
          free (tmpClient);
        }
        else {
          if (fcntl(tmpClient->fd,F_SETFL,O_NONBLOCK)==-1) {
            close (tmpClient->fd);
          }
          tmpClient->next=clientList; /* Client joins the club */
          clientList=tmpClient;
        }
      }
    }
  }
}
